@@ -1,7 +1,7 @@
 /*-
   decode.c -- low-level decompressor
 
-  Copyright (C) 2011, 2012, 2013 Mikolaj Izdebski
+  Copyright (C) 2011, 2012, 2013, 2014 Mikolaj Izdebski
 
   This file is part of lbzip2.
 
@@ -23,7 +23,6 @@
 #include <arpa/inet.h>          /* ntohl() */
 #include <string.h>             /* memcpy() */
 
-#include "main.h"
 #include "decode.h"
 
 
@@ -250,12 +249,12 @@ make_tree(struct retriever_internal_state *rs)
      greater than the greatest possible left-justified base).
    */
   assert(k == MAX_CODE_LENGTH + 1);
-  k = MAX_CODE_LENGTH;
-  while (C[k] == 0) {
+  do {
     assert(k > MIN_CODE_LENGTH);
-    assert(B[k] == 0);
+    assert(k > MAX_CODE_LENGTH || B[k] == 0);
     B[k--] = -1;
   }
+  while (C[k] == 0);
 
   /* Transform counts into indices (cumulative counts). */
   cum = 0;
@@ -456,7 +455,7 @@ mtf_one(uint8_t **imtf_row, uint8_t *imtf_slide, uint8_t c)
     switch (nn) {
     default:
       abort();
-#define R(n) case n: pp[n] = pp[n-1]
+#define R(n) case n: pp[n] = pp[n-1]; FALLTHROUGH
       R(15); R(14); R(13); R(12); R(11); R(10); R(9);
       R(8); R(7); R(6); R(5); R(4); R(3); R(2); R(1);
 #undef R
@@ -692,7 +691,7 @@ retrieve(struct decoder_state *restrict ds, struct bitstream *bs)
             continue;
           }
 
-          if (unlikely(run > tt_limit - tt)) {
+          if (unlikely(run > (size_t)(tt_limit - tt))) {
             return ERR_OVERFLOW;
           }
 
@@ -736,7 +735,7 @@ retrieve(struct decoder_state *restrict ds, struct bitstream *bs)
 
           if (unlikely(IS_EOB(s))) {
           eob:
-            if (unlikely(rs->run > tt_limit - tt))
+            if (unlikely(rs->run > (size_t)(tt_limit - tt)))
               return ERR_OVERFLOW;
 
             ds->ftab[rs->runChar] += rs->run;
@@ -752,8 +751,6 @@ retrieve(struct decoder_state *restrict ds, struct bitstream *bs)
             if (ds->bwt_idx >= ds->block_size)
               return ERR_BWTIDX;
 
-            free(ds->internal_state);
-            ds->internal_state = NULL;
             return OK;
           }
 
@@ -773,7 +770,7 @@ retrieve(struct decoder_state *restrict ds, struct bitstream *bs)
              bytes.  Zero-length run is possible only at the beginning,
              once per block, so any optimization involving zero-length
              runs are pointless. */
-          if (unlikely(rs->run > tt_limit - tt)) {
+          if (unlikely(rs->run > (size_t)(tt_limit - tt))) {
             return ERR_OVERFLOW;
           }
 
@@ -851,7 +848,7 @@ static const uint16_t rand_table[512] = {
 void
 decode(struct decoder_state *ds)
 {
-  uint32_t i, j = 0;
+  uint32_t i, j, k;
   uint32_t cum;
   uint8_t uc;
 
@@ -891,32 +888,34 @@ decode(struct decoder_state *ds)
      it usually leads to decreased compression ratio.
    */
   if (unlikely(ds->rand)) {
-    uint8_t *block;
-
-    /* Allocate a temporary array to hold the block. */
-    block = XNMALLOC(ds->block_size, uint8_t);
-
-    /* Copy the IBWT linked list into the temporary array. */
-    j = tt[ds->bwt_idx];
+    /* Compute IBWT in-situ.  A slower algorithm based on binary search is used
+       to avoid extra memory allocation. */
+    j = ds->bwt_idx;
     for (i = 0; i < ds->block_size; i++) {
-      j = tt[j >> 8];
-      block[i] = j;
+      k = 0;
+      if (j >= ds->ftab[k + 127]) k += 128;
+      if (j >= ds->ftab[k +  63]) k +=  64;
+      if (j >= ds->ftab[k +  31]) k +=  32;
+      if (j >= ds->ftab[k +  15]) k +=  16;
+      if (j >= ds->ftab[k +   7]) k +=   8;
+      if (j >= ds->ftab[k +   3]) k +=   4;
+      if (j >= ds->ftab[k +   1]) k +=   2;
+      if (j >= ds->ftab[k +   0]) k +=   1;
+      tt[i] = (tt[i] & ~0xFF) + k;
+      j = tt[j] >> 8;
     }
 
     /* Derandomize the block. */
     i = 0, j = RAND_THRESH;
     while (j < ds->block_size) {
-      block[j] ^= 1;
+      tt[j] ^= 1;
       i = (i + 1) & 0x1FF;
       j += rand_table[i];
     }
 
-    /* Reform a linked list from the array. */
+    /* Reform a linked list. */
     for (i = 0; i < ds->block_size; i++)
-      tt[i] = ((i + 1) << 8) + block[i];
-
-    /* Release the temporary array. */
-    free(block);
+      tt[i] = ((i + 1) << 8) + (tt[i] & 0xFF);
   }
 
   ds->rle_state = 0;
@@ -982,6 +981,7 @@ emit(struct decoder_state *ds, void *buf, size_t *buf_sz)
     if (unlikely(!a--))
       break;
     c = p = t[p >> 8];
+    /* fall-through */
   case 2:
     if (unlikely(!m--)) {
       ds->rle_state = 2;
@@ -993,6 +993,7 @@ emit(struct decoder_state *ds, void *buf, size_t *buf_sz)
     if (unlikely(!a--))
       break;
     c = p = t[p >> 8];
+    /* fall-through */
   case 3:
     if (unlikely(!m--)) {
       ds->rle_state = 3;
@@ -1004,6 +1005,7 @@ emit(struct decoder_state *ds, void *buf, size_t *buf_sz)
     if (unlikely(!a--))
       return ERR_RUNLEN;
     c = p = t[p >> 8];
+    /* fall-through */
   case 4:
     if (unlikely(m < c)) {
       c -= m;
@@ -1015,10 +1017,12 @@ emit(struct decoder_state *ds, void *buf, size_t *buf_sz)
     m -= c;
     while (c--)
       s = (s << 8) ^ crc_table[(s >> 24) ^ (*b++ = d)];
+    /* fall-through */
   case 0:
     if (unlikely(!a--))
       break;
     c = p = t[p >> 8];
+    /* fall-through */
   case 5:
     if (unlikely(!m--)) {
       ds->rle_state = 5;
@@ -1137,20 +1141,19 @@ emit(struct decoder_state *ds, void *buf, size_t *buf_sz)
 }
 
 
+size_t
+decoder_alloc_size(void)
+{
+  return (sizeof(struct decoder_state) +
+          MAX_BLOCK_SIZE * sizeof(uint32_t) +
+          sizeof(struct retriever_internal_state));
+}
+
+
 void
 decoder_init(struct decoder_state *ds)
 {
-  ds->internal_state = XMALLOC(struct retriever_internal_state);
+  ds->internal_state = (void *)(ds->tt + MAX_BLOCK_SIZE);
   ds->internal_state->state = S_INIT;
-
-  ds->tt = XNMALLOC(MAX_BLOCK_SIZE, uint32_t);
   ds->block_size = 0;
 }
-
-
-void
-decoder_free(struct decoder_state *ds)
-{
-  free(ds->tt);
-  free(ds->internal_state);
-}
