@@ -1,7 +1,7 @@
 /*-
   expand.c -- high-level decompression routines
 
-  Copyright (C) 2011, 2012 Mikolaj Izdebski
+  Copyright (C) 2011, 2012, 2014 Mikolaj Izdebski
 
   This file is part of lbzip2.
 
@@ -25,7 +25,7 @@
 #include "main.h"               /* bs100k */
 #include "process.h"            /* struct process */
 
-#include <strings.h>            /* bzero() */
+#include <string.h>             /* memset() */
 
 
 /*
@@ -119,13 +119,13 @@ struct unord_blk {
 struct retr_blk {
   struct detached_bitstream curr_pos;
   struct position base;
-  struct decoder_state ds;
+  struct decoder_state *ds;
   struct unord_blk *unord_link;
 };
 
 struct emit_blk {
   struct position base;
-  struct decoder_state ds;
+  struct decoder_state *ds;
   int status;
   uintmax_t end_offset;
 };
@@ -371,7 +371,7 @@ advance(struct detached_bitstream bs)
     Trace(("Advanced over miss-recognized bit pattern at {%u}",
            nbsx2(rb->base)));
 
-    decoder_free(&rb->ds);
+    free(rb->ds);
     free(rb);
     work_units++;
   }
@@ -456,7 +456,7 @@ do_parse(void)
       Trace(("Parser discovered a bit pattern beyond EOF at {%u}",
              nbsx2(rb->base)));
 
-      decoder_free(&rb->ds);
+      free(rb->ds);
       free(rb);
       work_units++;
     }
@@ -532,7 +532,8 @@ do_parse(void)
     struct retr_blk *rb = XMALLOC(struct retr_blk);
 
     rb->unord_link = NULL;
-    decoder_init(&rb->ds);
+    rb->ds = xmalloc(decoder_alloc_size());
+    decoder_init(rb->ds);
     rb->curr_pos = parser_bs;
     rb->base = parser_bs.pos;
     enqueue(retr_q, rb);
@@ -561,11 +562,11 @@ do_retrieve(void)
   rb = dequeue(retr_q);
 
   true_bitstream = attach(rb->curr_pos);
-  rv = retrieve(&rb->ds, &true_bitstream);
+  rv = retrieve(rb->ds, &true_bitstream);
   rb->curr_pos = detach(true_bitstream);
 
   if (parsing_done) {
-    decoder_free(&rb->ds);
+    free(rb->ds);
     free(rb);
     work_units++;
     check_invariants();
@@ -579,7 +580,7 @@ do_retrieve(void)
        abort this retrieve job. */
     Trace(("Retriever found himself redundand"));
     work_units++;
-    decoder_free(&rb->ds);
+    free(rb->ds);
     free(rb);
     check_invariants();
     return;
@@ -623,7 +624,7 @@ do_retrieve(void)
   sched_unlock();
 
   if (rv == OK)
-    decode(&rb->ds);
+    decode(rb->ds);
 
   eb = XMALLOC(struct emit_blk);
 
@@ -663,10 +664,10 @@ do_emit(void)
 
   oblk = xmalloc(sizeof(struct out_blk) + out_granul);
   oblk->size = out_granul;
-  oblk->blk_sz = eb->ds.block_size;
+  oblk->blk_sz = eb->ds->block_size;
   rv = eb->status;
   if (rv == OK)
-    rv = emit(&eb->ds, oblk + 1, &oblk->size);
+    rv = emit(eb->ds, oblk + 1, &oblk->size);
   oblk->size = out_granul - oblk->size;
   oblk->status = rv;
   oblk->base = eb->base;
@@ -679,8 +680,8 @@ do_emit(void)
   }
   else {
     oblk->end_offset = eb->end_offset;
-    oblk->crc = eb->ds.crc;
-    decoder_free(&eb->ds);
+    oblk->crc = eb->ds->crc;
+    free(eb->ds);
     free(eb);
     sched_lock();
     work_units++;
@@ -744,7 +745,7 @@ static bool
 can_scan(void)
 {
   return ((work_units > SCAN_THRESH || (work_units > 0u && !parse_token))
-          && !empty(scan_q) && can_attach(*peek(scan_q)));
+          && !ultra && !empty(scan_q) && can_attach(*peek(scan_q)));
 }
 
 static void
@@ -797,7 +798,8 @@ do_scan(void)
 
     rb = XMALLOC(struct retr_blk);
     rb->unord_link = ub;
-    decoder_init(&rb->ds);
+    rb->ds = xmalloc(decoder_alloc_size());
+    decoder_init(rb->ds);
     rb->curr_pos = *bs;
     rb->base = bs->pos;
     enqueue(retr_q, rb);
@@ -839,7 +841,7 @@ on_input_avail(void *buffer, size_t size)
      can possibly alter it, so here we can read tail_offs without locking. */
 
   missing = -size % 4;
-  bzero((char *)buffer + size, missing);
+  memset((char *)buffer + size, 0, missing);
 
   scan_task = XMALLOC(struct detached_bitstream);
   *scan_task = bits_init(tail_offs);
@@ -896,7 +898,7 @@ init(void)
   reord_offs = 0;
 
   parser_bs = bits_init(0);
-  parser_init(&par, bs100k);
+  parser_init(&par, bs100k, 0);
 }
 
 
