@@ -1,7 +1,7 @@
 /*-
   main.c -- main module
 
-  Copyright (C) 2011, 2012 Mikolaj Izdebski
+  Copyright (C) 2011, 2012, 2014 Mikolaj Izdebski
   Copyright (C) 2008, 2009, 2010 Laszlo Ersek
 
   This file is part of lbzip2.
@@ -32,6 +32,7 @@
 
 #include "stat-time.h"          /* get_stat_atime() */
 #include "utimens.h"            /* fdutimens() */
+#include "xalloc.h"             /* XMALLOC() */
 
 #include "signals.h"            /* setup_signals() */
 #include "main.h"               /* pname */
@@ -46,6 +47,7 @@ bool keep;                      /* -k */
 bool verbose;                   /* -v */
 bool print_cctrs;               /* -S */
 bool small;                     /* -s */
+bool ultra;                     /* -u */
 struct filespec ispec;
 struct filespec ospec;
 
@@ -76,7 +78,8 @@ cleanup(void)
 /* Called when one of xalloc functions fails. */
 void xalloc_die(void)
 {
-  failx(errno, "xalloc");
+  fail("Insufficient memory to complete operation."
+       " See manual page for ways of reducing memory usage.");
 }
 
 
@@ -169,11 +172,11 @@ xstrtol(const char *str, int source, uintmax_t lower, uintmax_t upper)
     goto fail;
   val = tmp;
 
-  endptr = index(suffix, *endptr);
+  endptr = strchr(suffix, *endptr);
   if (endptr == NULL)
     goto fail;
 
-  shift = (index(suffix, '\0') - endptr + 1) / 2 * 10;
+  shift = (strchr(suffix, '\0') - endptr + 1) / 2 * 10;
 
   if (val > (UINTMAX_MAX >> shift))
     goto fail;
@@ -218,25 +221,29 @@ G.\n  -z, --compress     : Force compression over the selection by PROG.\n  -1\
 he default.\n  -f, --force        : Open non-regular input files. Open input f\
 iles with more\n                       than one", " link. Try to remove each o\
 utput file before\n                       opening it. With `-cd' copy files no\
-t in bzip2 format.\n  -v, --verbose      : Log each (de)compression start to s\
-tderr. Display\n                       compression ratio and space savings. Di\
-splay progress\n                       information if stderr is connected to a\
- terminal.\n  -S                 : Print condition variable statistics to stde\
-rr.\n  -s, --small, -q,\n  --quiet,\n  --repetitive-fast,\n  --repetitive-best\
-,\n  --exponential      ", ": Accepted for compatibility, otherwise ignored.\n\
-  -h, --help         : Print this help to stdout and exit.\n  -L, --license, -\
-V,\n  --version          : Print version information to stdout and exit.\n\nOp\
-erands:\n\n  FILE               : Specify files to compress or decompress. If \
-no FILE is\n                       given, work as a filter. FILEs with `.bz2',\
- `.tbz',\n                       `.tbz2' and `.tz2' name suffixes will be skip\
-ped when\n                       compressing. When decompressing, `.bz2' suffi\
-x", "es will be\n                       removed in output filenames; `.tbz', `\
-.tbz2' and `.tz2'\n                       suffixes will be replaced by `.tar';\
- other filenames\n                       will be suffixed with `.out'.\n"
+t in bzip2 format.\n  -s, --small        : Reduce memory usage at cost of perf\
+ormance.\n  -u, --sequential   : Perform splitting input blocks sequentially. \
+This may\n                       improve compression ratio and decrease CPU us\
+age, but\n                       will degrade scalability.\n  -v, --verbose   \
+   : Log each (de)compression start to stderr. Display\n                      \
+ compression ratio an", "d space savings. Display progress\n                  \
+     information if stderr is connected to a terminal.\n  -S                 :\
+ Print condition variable statistics to stderr.\n  -q, --quiet,\n  --repetitiv\
+e-fast,\n  --repetitive-best,\n  --exponential      : Accepted for compatibili\
+ty, otherwise ignored.\n  -h, --help         : Print this help to stdout and e\
+xit.\n  -L, --license, -V,\n  --version          : Print version information t\
+o stdout and exit.\n\nOperands:\n\n  FILE               : Specify files to com\
+p", "ress or decompress. If no FILE is\n                       given, work as \
+a filter. FILEs with `.bz2', `.tbz',\n                       `.tbz2' and `.tz2\
+' name suffixes will be skipped when\n                       compressing. When\
+ decompressing, `.bz2' suffixes will be\n                       removed in out\
+put filenames; `.tbz', `.tbz2' and `.tz2'\n                       suffixes wil\
+l be replaced by `.tar'; other filenames\n                       will be suffi\
+xed with `.out'.\n"
 
 #define HELP_STRING "%s version %s\n%s\n\n%s%s",                        \
     PACKAGE_NAME, PACKAGE_VERSION, "http://lbzip2.org/",                \
-    "Copyright (C) 2011, 2012 Mikolaj Izdebski\n"                       \
+    "Copyright (C) 2011, 2012, 2013, 2014 Mikolaj Izdebski\n"           \
     "Copyright (C) 2008, 2009, 2010 Laszlo Ersek\n"                     \
     "\n"                                                                \
     "This program is free software: you can redistribute it and/or modify\n" \
@@ -253,24 +260,26 @@ x", "es will be\n                       removed in output filenames; `.tbz', `\
     "ong with this program.  If not, see <http://www.gnu.org/licenses/>.\n"
 
 
-static void _Noreturn
+_Noreturn static void
 usage(void)
 {
   if (0 > printf(USAGE_STRING))
     failx(errno, "printf()");
   if (0 != fclose(stdout))
     failx(errno, "fclose(stdout)");
+  gcov_flush();
   _exit(EX_OK);
 }
 
 
-static void _Noreturn
+_Noreturn static void
 version(void)
 {
   if (0 > printf(HELP_STRING))
     failx(errno, "printf()");
   if (0 != fclose(stdout))
     failx(errno, "fclose(stdout)");
+  gcov_flush();
   _exit(EX_OK);
 }
 
@@ -432,6 +441,9 @@ opts_setup(struct arg **operands, size_t argc, char **argv)
           else if (0 == strcmp("small", argscan)) {
             small = 1;
           }
+          else if (0 == strcmp("sequential", argscan)) {
+            ultra = 1;
+          }
           else if (0 == strcmp("verbose", argscan)) {
             verbose = 1;
           }
@@ -496,6 +508,10 @@ opts_setup(struct arg **operands, size_t argc, char **argv)
               small = 1;
               break;
 
+            case 'u':
+              ultra = 1;
+              break;
+
             case 'v':
               verbose = 1;
               break;
@@ -965,5 +981,6 @@ main(int argc, char **argv)
     failx(errno, "close(stdout)");
   }
 
+  gcov_flush();
   _exit(warned ? EX_WARN : EX_OK);
 }
